{% load static %}

<meta name="reactor-boost" data-enabled="{{ BOOST_PAGES|lower }}"/>
{% for c in components %}
  <meta name="reactor-component"
        data-tag-name="{{ c.tag_name }}"
        data-extends="{{ c.extends }}"/>
{% endfor %}
<style>
  .reactor-disconnected {opacity: 0.8; cursor: wait;}
  .reactor-disconnected * {opacity: 0.8; pointer-events: none;}
</style>
<script defer src="{% static "reactor/reactor.min.js" %}"></script>
